"""
A package that models a generic load-bearing wall for the purpose of determining
reactions and load-transfer.

The wall model is parameterizable and can represent any material
"""

__version__ = "0.4.1"

from loadbearing_wall.wall_model import LinearWallModel
from loadbearing_wall import *
